unit xazones;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  CommCtrl, ComCtrls, StdCtrls, ExtCtrls, Math, DsgnIntf;

type
  // -- TXAZoneControl ----------------------------------------------
  TXAZoneControl = class;

  // -- TXAZoneArea -------------------------------------------------
  TXAZoneArea = class( TCustomControl )
    private    // -- Private Declarations -------
      FActive                : Boolean;
      FOnHide                : TNotifyEvent;
      FOnShow                : TNotifyEvent;
      function GetTabIndex : Integer;
      procedure SetActive( Value: Boolean );
      procedure SetTabIndex( Value: Integer );
    protected  // -- Protected Declarations -----
      procedure CreateParams( var Params: TCreateParams ); override;
      procedure CMTextChanged( var Msg: TMessage ); message CM_TEXTCHANGED;
      procedure Paint; override;
      function ParentAsZC : TXAZoneControl;
      function ParentIsZC : Boolean;
    public     // -- Public Declarations --------
      constructor Create( AOwner: TComponent ); override;
      destructor Destroy; override;
      property Active          : Boolean            read FActive         write SetActive;
    published  // -- Published Declarations -----
      property Caption;
      property Color;
      property Enabled;
      property TabIndex        : Integer            read GetTabIndex     write SetTabIndex;
      property OnHide          : TNotifyEvent       read FOnHide         write FOnHide;
      property OnShow          : TNotifyEvent       read FOnShow         write FOnShow;
   end;

  // -- TXAZoneControl ----------------------------------------------
  TXAZoneControlEditor = class( TComponentEditor )
    public     // -- Public Declarations --------
      function GetVerbCount : Integer; override;
      function GetVerb( Index: Integer ) : string; override;
      procedure ExecuteVerb( Index: Integer ); override;
   end;

  // -- TXAZoneControl Events ---------------------------------------
  TXAZCChangingEvent = procedure ( Sender: TObject; var AllowChange: Boolean ) of object;
  TXAZoneButtonEvent = procedure ( Sender: TObject; Tab: Integer ) of object;

  // -- TXAZoneControl - v0.9.3 -------------------------------------
  TXAZoneControl = class( TCustomControl )
    private    // -- Private Declarations -------
      FFrameWidth            : Integer;
      FTabColor              : TColor;
      FTabHeight             : Integer;
      FOnChange              : TNotifyEvent;
      FOnChanging            : TXAZCChangingEvent;
      procedure Changer( AZone: TXAZoneArea );
      function GetActiveZone : TXAZoneArea;
      function GetZACount : Integer;
      function GetZAAtPos( APoint: TPoint ) : TXAZoneArea;
      function GetDisplayRect : TRect;
      function GetTabRect( AIndex: Integer ) : TRect;
      function GetZoneRect : TRect;
      procedure SetActiveZone( AZone: TXAZoneArea );
      procedure SetFrameWidth( AValue: Integer );
      procedure SetTabColor( AValue: TColor );
      procedure SetTabHeight( AValue: Integer );
      procedure ResizeZones;
    protected  // -- Protected Declarations -----
      procedure Notification( AComponent: TComponent; Operation: TOperation ); override;
      procedure AdjustClientRect( var Rect: TRect ); override;
      procedure WMGetDlgCode( var Msg: TWMGetDlgCode ); message WM_GETDLGCODE;
      procedure CMDesignHitTest( var Msg: TCMDesignHitTest ); message CM_DESIGNHITTEST;
      procedure CMFontChanged( var Msg: TMessage ); message CM_FONTCHANGED;
      procedure MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer ); override;
      procedure KeyDown( var Key: Word; Shift: TShiftState ); override;
      procedure Change;
      procedure ClickAtPos( APoint: TPoint );
      procedure Paint; override;
    public     // -- Public Declarations --------
      constructor Create( AOwner: TComponent ); override;
      destructor Destroy; override;
      function ChildAsZA( AControl: TControl ) : TXAZoneArea;
      function ChildIsZA( AControl: TControl ) : Boolean;
      function FindNextTab( AZone: TXAZoneArea ) : TXAZoneArea;
      function FindPrevTab( AZone: TXAZoneArea ) : TXAZoneArea;
    published  // -- Published Declarations -----
      property Align;
      property ActiveZone      : TXAZoneArea        read GetActiveZone   write SetActiveZone;
      property Anchors;
      property Color;
      property Constraints;
      property Ctl3D;
      property ZACount         : Integer            read GetZACount;
      property Font;
      property FrameWidth      : Integer            read FFrameWidth     write SetFrameWidth;
      property TabColor        : TColor             read FTabColor       write SetTabColor;
      property TabHeight       : Integer            read FTabHeight      write SetTabHeight;
      property Parent;
      property PopupMenu;
      property TabOrder;
      property TabStop;
      property Visible;
      property OnChange        : TNotifyEvent       read FOnChange       write FOnChange;
      property OnChanging      : TXAZCChangingEvent read FOnChanging     write FOnChanging;
      property OnClick;                      // click events
      property OnDblClick;
      property OnDragOver;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMouseDown;                  // mouse events
      property OnMouseMove;
      property OnMouseUp;
   end;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Extras', [TXAZoneControl]);
  RegisterNoIcon([TXAZoneArea]);
  RegisterClass(TXAZoneArea);
  RegisterComponentEditor(TXAZoneControl, TXAZoneControlEditor);
  RegisterComponentEditor(TXAZoneArea, TXAZoneControlEditor);
end;

// ------------------------------------------------------------------
// -- TXAZoneArea ---------------------------------------------------
// ------------------------------------------------------------------
constructor TXAZoneArea.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  Self.BoundsRect := Rect( 10,10,20,20 );
  Self.BorderWidth := 1;
  Self.ControlStyle := [ csAcceptsControls, csCaptureMouse, csClickEvents,
                         csSetCaption, csOpaque, csDoubleClicks,
                         csReplicatable ];
  Self.Visible := True;
  Self.Color := clTeal;
end;

destructor TXAZoneArea.Destroy;
begin
  inherited Destroy;
end;

procedure TXAZoneArea.CreateParams( var Params: TCreateParams );
begin
  inherited CreateParams(Params);
  with Params do
  begin
    if (csDesigning in ComponentState) then  // -- Display Border when
         Style := Style or WS_BORDER         // -- in Design Mode
    else Style := Style or 0;
    WindowClass.Style := WindowClass.Style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

// --------------------------------------------------------
// -- Capture Changes to the Display Caption for the Zone
procedure TXAZoneArea.CMTextChanged( var Msg: TMessage );
begin
  inherited;
  if Self.ParentIsZC then
    Self.ParentAsZC.Invalidate;
end;

// --------------------------------------------------------
function TXAZoneArea.GetTabIndex : Integer;
begin
  Result := Self.TabOrder;
end;

procedure TXAZoneArea.SetActive( Value: Boolean );
begin
  if (Value <> FActive) then
  begin
    FActive := Value;
    case Active of
     True : if Assigned(FOnShow) then FOnShow( Self );
       else if Assigned(FOnHide) then FOnHide( Self );
    end;
  end;
end;

procedure TXAZoneArea.SetTabIndex( Value: Integer );
begin
  if Self.ParentIsZC then
  if (Value >= 0) and (Value <= (Self.ParentAsZC.ZACount-1)) then
  begin
    Self.TabOrder := Value;
    Self.ParentAsZC.ResizeZones;
    Self.ParentAsZC.Invalidate;
  end;
end;

// --------------------------------------------------------
procedure TXAZoneArea.Paint;
begin
  inherited;
  with Canvas do
  begin
    Brush.Color := Self.Color;
    FillRect( Self.ClientRect );
  end;
end;

// --------------------------------------------------------
function TXAZoneArea.ParentAsZC : TXAZoneControl;
begin
  Result := TXAZoneControl( Self.Parent );
end;

function TXAZoneArea.ParentIsZC : Boolean;
begin
  Result := (Self.Parent <> nil) and (Self.Parent is TXAZoneControl);
end;


// ------------------------------------------------------------------
// -- TXAZoneControl Contextual Menu Editor Controller --------------
// ------------------------------------------------------------------
function TXAZoneControlEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

function TXAZoneControlEditor.GetVerb( Index: Integer ) : string;
begin
  case Index of
   0 : Result := 'New &Area';
  end;
end;

procedure TXAZoneControlEditor.ExecuteVerb( Index: Integer );
var
  ovDesigner                 : IFormDesigner;
  ovZoneControl              : TXAZoneControl;
  ovNewZone                  : TXAZoneArea;
begin
  ovDesigner := Self.Designer;
  case Index of
   0 : begin   // -- New Area -------------------
         if (Component is TXAZoneControl) then
              ovZoneControl := (Component as TXAZoneControl)
         else ovZoneControl := (TXAZoneArea(Component).Parent as TXAZoneControl);
         ovNewZone := TXAZoneArea.Create( ovDesigner.Form );
         ovNewZone.Name := ovDesigner.UniqueName( TXAZoneArea.ClassName );
         ovNewZone.Caption := ovNewZone.Name;
         ovNewZone.Parent := ovZoneControl;
         ovZoneControl.ActiveZone := ovNewZone;
         ovDesigner.SelectComponent( ovNewZone );
       end;
  end;
  Designer.Modified;
end;


// ------------------------------------------------------------------
// -- TXAZoneControl ------------------------------------------------
// ------------------------------------------------------------------
constructor TXAZoneControl.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FFrameWidth := 1;
  Self.BoundsRect := Rect( 10,10,60,110 );
  Self.Align := alLeft;
  Self.ControlStyle := [ csCaptureMouse, csClickEvents, csOpaque,
                         csDoubleClicks ];
  Self.Color := clGray;
  Self.TabStop := True;
  FTabColor := clSilver;
  FTabHeight := 14;
end;

destructor TXAZoneControl.Destroy;
begin
  inherited Destroy;
end;

// --------------------------------------------------------
procedure TXAZoneControl.Notification( AComponent: TComponent; Operation: TOperation );
begin
  inherited Notification( AComponent, Operation );
  if (AComponent is TXAZoneArea) then
  case Operation of
   opInsert : SetActiveZone( TXAZoneArea(AComponent) );
   opRemove : if (Self.ControlCount <= 0) then
                   SetActiveZone( nil )
              else if ChildIsZA( Self.Controls[0] ) then
                     SetActiveZone( ChildAsZA( Self.Controls[0] ) );
  end;
end;

procedure TXAZoneControl.AdjustClientRect( var Rect: TRect );
begin
  Rect := GetDisplayRect;
  InflateRect( Rect, -FFrameWidth, -FFrameWidth );
  InflateRect( Rect, 0, -100 );
  inherited AdjustClientRect( Rect );
end;

// -- Sets the component to TRAP the requested element types
procedure TXAZoneControl.WMGetDlgCode( var Msg: TWMGetDlgCode );
begin   // _WANTARROWS,  _WANTTAB,  _WANTCHARS
  Msg.Result := (DLGC_WANTARROWS + DLGC_WANTCHARS);
end;

// -- Captures Mouse movement and Clicking during Design-Time
procedure TXAZoneControl.CMDesignHitTest( var Msg: TCMDesignHitTest );
const
  ovControlClick = 1;
begin
  inherited;
  if (Msg.Keys = ovControlClick) then
    ClickAtPos( Point( Msg.Pos.x, Msg.Pos.y ) );
end;

// -- Captures Changes to the Display Font
procedure TXAZoneControl.CMFontChanged( var Msg: TMessage );
begin
  inherited;
  Invalidate;
end;

procedure TXAZoneControl.MouseDown( Button: TMouseButton; Shift: TShiftState;
                                    X, Y: Integer );
begin
  inherited;
  if (not (csDesigning in Self.ComponentState)) and
     (Self.CanFocus or (GetParentForm(Self) = nil)) then
  begin
    Self.SetFocus;
    if (not Self.Focused) then Exit;
  end;
  ClickAtPos( Point( X, Y ) );
end;

procedure TXAZoneControl.KeyDown( var Key: Word; Shift: TShiftState );
begin
  inherited KeyDown(Key, Shift);
  if (Key in [VK_UP,VK_DOWN]) then
  begin
//    case Key of
//       VK_UP : ActiveZone := FindPrevTab( ActiveZone );
//     VK_DOWN : ActiveZone := FindNextTab( ActiveZone );
//    end;
  end;
end;

// --------------------------------------------------------
procedure TXAZoneControl.Change;
begin
  inherited;
  Changer( Self.ActiveZone );
end;

procedure TXAZoneControl.Changer( AZone: TXAZoneArea );
var
  ovForm                     : TCustomForm;
  ovChangeOK, ovInDesign     : Boolean;
begin
  ovForm := GetParentForm( Self );     //-- D/T Property/Component Editors Link
  ovChangeOK := True;                  //-- State Processing Switch Definition
  ovInDesign := (csDesigning in Self.ComponentState);
  if (not ovInDesign) and Assigned(FOnChanging) then
    FOnChanging( Self, ovChangeOK );   //-- Check validity of Zone Change
  if ovChangeOK then
  begin
    if (AZone <> nil) then SetActiveZone( AZone )      //-- Update Active Zone
                      else SetActiveZone( nil );
    if ovInDesign and (ovForm <> nil) then             //-- Update D/T Editors
      if (ovForm.Designer <> nil) then ovForm.Designer.Modified;
    if AZone.Active and (AZone = Self.ActiveZone) then //-- Display Active Zone
         AZone.BringToFront
    else AZone.SendToBack;
    if (not ovInDesign) and Assigned(FOnChange) then
      FOnChange( Self );               //-- Notify User of Zone Change
  end;
end;

procedure TXAZoneControl.ClickAtPos( APoint: TPoint );
var
  ovZone                     : TXAZoneArea;
begin
  ovZone := GetZAAtPos( APoint );
  if (ovZone <> nil) then
  if (ActiveZone = nil) or (ActiveZone <> ovZone) then
    Changer( ovZone );
end;

// --------------------------------------------------------
function TXAZoneControl.GetActiveZone : TXAZoneArea;
var
  ovZone                     : TXAZoneArea;
  ovLoop                     : Integer;
begin
  Result := nil;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    if ovZone.Active then Result := ovZone;
  end;
end;

function TXAZoneControl.GetZACount : Integer;
var
  ovLoop                     : Integer;
begin
  Result := 0;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
    Inc( Result );
end;

function TXAZoneControl.GetZAAtPos( APoint: TPoint ) : TXAZoneArea;
var
  ovZone                     : TXAZoneArea;
  ovLoop                     : Integer;
begin
  Result := nil;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    if PtInRect(GetTabRect(ovZone.TabIndex), APoint) then Result := ovZone;
  end;
end;

function TXAZoneControl.GetDisplayRect : TRect;
begin
  Result := Self.ClientRect;
end;

function TXAZoneControl.GetTabRect( AIndex: Integer ) : TRect;
begin
  if (AIndex < 0) or (AIndex > (ZACount-1)) then
       Result := Rect( 0,0,0,0 )
  else begin
         Result := Rect( 2, 0, Width-2, 1 );
         if (AIndex <= ActiveZone.TabIndex) then
              Result.Top := ((AIndex * TabHeight) + 1)
         else Result.Top := (Height - ((ZACount - AIndex) * TabHeight) - 1);
         Result.Bottom := (Result.Top + TabHeight - 1);
       end;
end;

function TXAZoneControl.GetZoneRect : TRect;
var
  ovActiveIndex              : Integer;
begin
  if (ActiveZone = nil) then
       ovActiveIndex := 0
  else ovActiveIndex := ActiveZone.TabIndex;
  if (ovActiveIndex < 0) or (ovActiveIndex > (ZACount-1)) then
       Result := Rect( 0,0,0,0 )
  else begin
         Inc( ovActiveIndex );
         Result := Rect( 1, 0, (Width - 2), 0 );
         Result.Top := (ovActiveIndex * TabHeight);
         Result.Bottom := (Height-(Max(0,(ZACount-ovActiveIndex)) * TabHeight)-1);
       end;
end;

procedure TXAZoneControl.SetActiveZone( AZone: TXAZoneArea );
var
  ovZone                     : TXAZoneArea;
  ovLoop                     : Integer;
begin
  for ovLoop := (Self.ControlCount-1) downto 0 do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    ovZone.Active := (AZone = ovZone);
  end;
  Self.Invalidate;
end;

procedure TXAZoneControl.SetFrameWidth( AValue: Integer );
begin
  if (AValue <> FFrameWidth) and (AValue in [0..10]) then
  begin
    FFrameWidth := AValue;
    Self.Invalidate;
  end;
end;

procedure TXAZoneControl.SetTabColor( AValue: TColor );
begin
  if (AValue <> FTabColor) then
  begin
    FTabColor := AValue;
    Self.Invalidate;
  end;
end;

procedure TXAZoneControl.SetTabHeight( AValue: Integer );
begin
  if (AValue <> FTabHeight) and (AValue in [0..50]) then
  begin
    FTabHeight := AValue;
    Self.Invalidate;
  end;
end;

// --------------------------------------------------------
procedure TXAZoneControl.Paint;
var
  ovZone                     : TXAZoneArea;
  ovRect                     : TRect;
  ovSize                     : TSize;
  ovLoop                     : Integer;
  ovCaption                  : string;
begin
  inherited;
  ResizeZones;
  ovRect := Self.ClientRect;
  with Self.Canvas do
  begin
    Font := Self.Font;
    Pen.Style := psSolid;
    Pen.Width := 1;
    Pen.Color :=  clBlack;
    Brush.Color := FTabColor;
    if (Self.ControlCount = 0) then FillRect( ovRect );
  end;
  Frame3D( Self.Canvas, ovRect, clBlack, clBtnHighlight, FFrameWidth );
  with Self.Canvas do
    for ovLoop := (Self.ControlCount-1) downto 0 do
    if ChildIsZA( Self.Controls[ovLoop] ) then
    begin
      ovZone := ChildAsZA( Self.Controls[ovLoop] );
      ovRect := GetTabRect( ovZone.TabIndex );
      FillRect( ovRect );
      ovCaption := ovZone.Caption;
      ovSize := TextExtent( ovCaption );
      TextOut( (Self.Width-3-ovSize.cx) div 2,
               ovRect.Top+((Self.TabHeight-ovSize.cy) div 2), ovCaption );
      Frame3D( Self.Canvas, ovRect, clBtnHighlight, clBlack, 1 );
    end;
end;

// --------------------------------------------------------
procedure TXAZoneControl.ResizeZones;
var
  ovRect                     : TRect;
  ovLoop                     : Integer;
begin
  ovRect := GetZoneRect;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
    ChildAsZA( Self.Controls[ovLoop] ).BoundsRect := ovRect;
end;

// --------------------------------------------------------
function TXAZoneControl.FindNextTab( AZone: TXAZoneArea ) : TXAZoneArea;
var
  ovLoop, ovCount, ovFindTab : Integer;
begin
  ovCount := Self.ZACount;
  if (ovCount < 2) or (AZone = nil) then
       Result := Self.ActiveZone
  else begin
         Result := nil;
         ovFindTab := AZone.TabIndex + 1;
         if (ovFindTab >= ovCount) then ovFindTab := 0;
         for ovLoop := 0 to (Self.ControlCount-1) do
         if ChildIsZA( Self.Controls[ovLoop] ) then
         if (ChildAsZA( Self.Controls[ovLoop] ).TabIndex = ovFindTab) then
           Result := ChildAsZA( Self.Controls[ovLoop] );
       end;
end;

function TXAZoneControl.FindPrevTab( AZone: TXAZoneArea ) : TXAZoneArea;
var
  ovLoop, ovCount, ovFindTab : Integer;
begin
  ovCount := Self.ZACount;
  if (ovCount < 2) or (AZone = nil) then
       Result := Self.ActiveZone
  else begin
         Result := nil;
         ovFindTab := AZone.TabIndex - 1;
         if (ovFindTab < 0) then ovFindTab := ovCount - 1;
         for ovLoop := 0 to (Self.ControlCount-1) do
         if ChildIsZA( Self.Controls[ovLoop] ) then
         if (ChildAsZA( Self.Controls[ovLoop] ).TabIndex = ovFindTab) then
           Result := ChildAsZA( Self.Controls[ovLoop] );
       end;
end;

// --------------------------------------------------------
function TXAZoneControl.ChildAsZA( AControl: TControl ) : TXAZoneArea;
begin
  Result := TXAZoneArea( AControl );
end;

function TXAZoneControl.ChildIsZA( AControl: TControl ) : Boolean;
begin
  Result := (AControl <> nil) and (AControl is TXAZoneArea);
end;

// ------------------------------------------------------------------

end.
